#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

struct dns_header {
  uint16_t id;
  uint16_t flags;
#define DNS_FLAG_QR (1 << 15)
#define DNS_FLAG_AA (1 << 10)
#define DNS_FLAG_TC (1 <<  9)
#define DNS_FLAG_RD (1 <<  8)
#define DNS_FLAG_RA (1 <<  7)
#define DNS_FLAG_AD (1 <<  5)
#define DNS_FLAG_CD (1 <<  4)
#define DNS_CODE_OPCODE_SHIFT  11
#define DNS_CODE_OPCODE_MASK  0xf
#define DNS_CODE_RCODE_SHIFT    0
#define DNS_CODE_RCODE_MASK   0xf
  uint16_t qdcount;
  uint16_t ancount;
  uint16_t nscount;
  uint16_t arcount;
};
#define DNS_HEADER_SIZE 12

struct dns_question {
  uint16_t type;
  uint16_t class;
};
#define DNS_QUESTION_SIZE 4

struct dns_rr {
  uint16_t type;
  uint16_t class;
  uint32_t ttl;
  uint16_t rdlength;
};
#define DNS_RR_SIZE 10

static unsigned char *write_name(struct dns_header *dnshdr, unsigned char *p,
				 int size, char *name)
{
  char *np = name, *pp;

  while (1) {
    pp = strchr(np, '.');
    if (!pp)
      pp = np + strlen(np);
    if (size < 1 + (pp - np) + 1)
      return NULL;
    *(p++) = pp - np;
    size--;
    memcpy(p, np, pp - np);
    p += pp - np;
    size -= pp - np;
    if (*pp == '\0')
      break;
    np = pp + 1;
  }
  *(p++) = 0x00;

  return p;
}

static int make_dns_request(unsigned char *buffer, int size, char *hostname,
			    int type)
{
  struct dns_header *dnshdr;
  struct dns_header header;
  struct dns_question question;
  unsigned char *p = buffer;

  header.id = htons(0x1234);
  header.flags = htons(DNS_FLAG_RD);

  header.qdcount = htons(1);
  header.ancount = 0;
  header.nscount = 0;
  header.arcount = 0;

  question.type  = htons(type);
  question.class = htons(1);

  dnshdr = (struct dns_header *)buffer;

  memcpy(p, &header, DNS_HEADER_SIZE);
  p += DNS_HEADER_SIZE;

  p = write_name(dnshdr, p, size - (p - buffer), hostname);
  if (!p)
    return -1;

  memcpy(p, &question, DNS_QUESTION_SIZE);
  p += DNS_QUESTION_SIZE;

  return p - buffer;
}

static unsigned char *read_name(struct dns_header *dnshdr, unsigned char *p,
				char *name, int size)
{
  int s, count = 0;
  char *np = name;
  unsigned char *end = NULL;

  while ((s = *(p++)) != 0) {
    if ((s & 0xc0) == 0xc0) {
      s = ((s & ~0xc0) << 8) + *(p++);
      if (!end) end = p;
      p = (unsigned char *)dnshdr + s;
      continue;
    }
    if (np - name + s + (count ? 1 : 0) + 1 > size)
      return NULL;
    if (count++) *(np++) = '.';
    memcpy(np, p, s);
    np += s;
    p += s;
  }
  *np = '\0';
  if (!end) end = p;

  return end;
}

static unsigned char *analyze_dns_question(struct dns_header *dnshdr,
					   unsigned char *p)
{
  char name[256];
  struct dns_question question;

  p = read_name(dnshdr, p, name, sizeof(name));
  if (!p) return NULL;
  printf("\t\tname\t: %s\n", name);
  memcpy(&question, p, DNS_QUESTION_SIZE);
  p += DNS_QUESTION_SIZE;
  printf("\t\ttype\t: %d\n", ntohs(question.type));
  printf("\t\tclass\t: %d\n", ntohs(question.class));

  return p;
}

static unsigned char *analyze_dns_rr(struct dns_header *dnshdr,
				     unsigned char *p)
{
  char name[256];
  struct dns_rr rr;
  int i;
  struct in_addr in;
  char in6_str[INET6_ADDRSTRLEN];
  uint16_t priority;

  p = read_name(dnshdr, p, name, sizeof(name));
  if (!p) return NULL;
  printf("\t\tname\t: %s\n", name);
  memcpy(&rr, p, DNS_RR_SIZE);
  p += DNS_RR_SIZE;
  printf("\t\ttype\t: %d\n", ntohs(rr.type));
  printf("\t\tclass\t: %d\n", ntohs(rr.class));
  printf("\t\tttl\t: %d\n", ntohl(rr.ttl));
  printf("\t\trdlen\t: %d\n", ntohs(rr.rdlength));

  printf("\t\trdata\t:");
  switch (ntohs(rr.type)) {
  case 1: /* A (IPv4 address) */
    memcpy(&in.s_addr, p, sizeof(in.s_addr));
    printf(" %s", inet_ntoa(in));
    break;
  case 2: /* NS */
    if (read_name(dnshdr, p, name, sizeof(name)))
      printf(" %s", name);
    break;
  case 15: /* MX */
    memcpy(&priority, p, sizeof(priority));
    priority = ntohs(priority);
    if (read_name(dnshdr, p + sizeof(priority), name, sizeof(name)))
      printf(" %s (%d)", name, (int)priority);
    break;
  case 28: /* IPv6 */
    printf(" %s", inet_ntop(AF_INET6, p, in6_str, sizeof(struct in6_addr)));
    break;
  default:
    for (i = 0; i < ntohs(rr.rdlength); i++)
      printf(" %02x", p[i]);
    break;
  }
  printf("\n");
  p += ntohs(rr.rdlength);

  return p;
}

static int analyze_dns(struct dns_header *dnshdr)
{
  unsigned char *p;
  int i;

  printf("DNS:\n");

  printf("\tID\t: %04x\n", ntohs(dnshdr->id));

  printf("\tFLAGS\t:");
  if (ntohs(dnshdr->flags) & DNS_FLAG_QR) printf(" QR");
  if (ntohs(dnshdr->flags) & DNS_FLAG_AA) printf(" AA");
  if (ntohs(dnshdr->flags) & DNS_FLAG_TC) printf(" TC");
  if (ntohs(dnshdr->flags) & DNS_FLAG_RD) printf(" RD");
  if (ntohs(dnshdr->flags) & DNS_FLAG_RA) printf(" RA");
  if (ntohs(dnshdr->flags) & DNS_FLAG_AD) printf(" AD");
  if (ntohs(dnshdr->flags) & DNS_FLAG_CD) printf(" CD");
  printf("\n");

  printf("\tOPCODE\t: %d\n",
	 (ntohs(dnshdr->flags) >> DNS_CODE_OPCODE_SHIFT)
	 & DNS_CODE_OPCODE_MASK);
  printf("\tRCODE\t: %d\n",
	 (ntohs(dnshdr->flags) >> DNS_CODE_RCODE_SHIFT)
	 & DNS_CODE_RCODE_MASK);

  printf("\tQDCOUNT\t: %d\n", ntohs(dnshdr->qdcount));
  printf("\tANCOUNT\t: %d\n", ntohs(dnshdr->ancount));
  printf("\tNSCOUNT\t: %d\n", ntohs(dnshdr->nscount));
  printf("\tARCOUNT\t: %d\n", ntohs(dnshdr->arcount));

  p = (unsigned char *)dnshdr + DNS_HEADER_SIZE;

  for (i = 0; i < ntohs(dnshdr->qdcount); i++) {
    printf("\tQuestion: (%d)\n", i);
    p = analyze_dns_question(dnshdr, p);
    if (!p) return -1;
  }

  for (i = 0; i < ntohs(dnshdr->ancount); i++) {
    printf("\tAnswer: (%d)\n", i);
    p = analyze_dns_rr(dnshdr, p);
    if (!p) return -1;
  }

  for (i = 0; i < ntohs(dnshdr->nscount); i++) {
    printf("\tAuthority: (%d)\n", i);
    p = analyze_dns_rr(dnshdr, p);
    if (!p) return -1;
  }

  for (i = 0; i < ntohs(dnshdr->arcount); i++) {
    printf("\tAdditional: (%d)\n", i);
    p = analyze_dns_rr(dnshdr, p);
    if (!p) return -1;
  }

  return 0;
}

int main(int argc, char *argv[])
{
  int sockt, port, size, type = 1;
  in_addr_t ipaddr;
  unsigned char buffer[4096];
  struct sockaddr_in addr;
  socklen_t addrlen;

  ipaddr = ntohl(inet_addr(argv[1]));
  port = atoi(argv[2]);

  sockt = socket(PF_INET, SOCK_DGRAM, 0);

  if (argc > 4) {
    if (!strcmp(argv[4], "A")) type = 1;
    else if (!strcmp(argv[4], "MX")) type = 15;
  }

  size = make_dns_request(buffer, sizeof(buffer), argv[3], type);
  if (size < 0) {
    fprintf(stderr, "Cannot make DNS request.\n");
    exit(1);
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(ipaddr);
  addrlen = sizeof(addr);

  sendto(sockt, buffer, size, 0, (struct sockaddr *)&addr, addrlen);

  memset(&addr, 0, sizeof(addr));
  addrlen = sizeof(addr);
  recvfrom(sockt, buffer, sizeof(buffer), 0,
	   (struct sockaddr *)&addr, &addrlen);

  analyze_dns((struct dns_header *)buffer);

  return 0;
}
